#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define VPCALC_SECTION	"MEM"
#define VPCALC_SUBSECT	"86"

#include "../config.h"
#include "../include/types.h"
#include "../include/debug.h"

#include "../include/memory.h"

#define PAGE(p)		((p) << 14)
#define PAGE_SIZE	(1 << 14)

typedef struct _memory_o {
	uint8 memory[0x4000 * 0x18];

	uint pages[4];
	uint8* rblocks[4];
	uint8* wblocks[4];
} _memory_o;

/* It doesn't matter if this is global, because it's write only memory.
 *   In fact, by making it global, we ensure that all the memory units will
 *   share the same scratch space.
 */
static uint8 scratch[0x4000];

memory_o memory_new(void)
{
	_memory_o *new;

	new = (_memory_o *) malloc(sizeof(_memory_o));

	if (new != NULL) {
		memset(new, 0, sizeof(_memory_o));

                new->pages[0] = 0x00;
                new->pages[1] = 0x01;
                new->pages[2] = 0x00;
                new->pages[3] = 0x10;
                                                                                
                new->rblocks[0] = new->memory + PAGE(new->pages[0]);
                new->rblocks[1] = new->memory + PAGE(new->pages[1]) - PAGE(1);
                new->rblocks[2] = new->memory + PAGE(new->pages[2]) - PAGE(2);
                new->rblocks[3] = new->memory + PAGE(new->pages[3]) - PAGE(3);
                                                                                
                new->wblocks[0] = scratch;
                new->wblocks[1] = scratch - PAGE(1);
                new->wblocks[2] = scratch - PAGE(2);
                new->wblocks[3] = new->rblocks[3];
	}

	return (memory_o)new;
}

void memory_del(memory_o memory)
{
	_memory_o *_memory = (_memory_o *)memory;

	free(_memory);
}

size_t memory_size(memory_o memory)
{
	_memory_o *_memory = (_memory_o *)memory;

	return sizeof(_memory->memory);
}

size_t memory_bus_size(memory_o memory)
{
	return 8;
}

size_t memory_addr_size(memory_o memory)
{
	return 16;
}

void memory_putimage(memory_o memory, void *data)
{
	_memory_o *_memory = (_memory_o *)memory;

	memcpy(_memory->memory, data, sizeof(_memory->memory));
}

void memory_getimage(memory_o memory, void *data)
{
	_memory_o *_memory = (_memory_o *)memory;

	memcpy(data, _memory->memory, sizeof(_memory->memory));
}

uint memory_fetch(memory_o memory, uint addr)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	return _memory->rblocks[addr >> 14][addr];
}

uint memory_read(memory_o memory, uint addr)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	return _memory->rblocks[addr >> 14][addr];
}

void memory_write(memory_o memory, uint addr, uint data)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	_memory->wblocks[addr >> 14][addr] = data;
}

uint8 memory_fetch_byte(memory_o memory, uint addr)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	return _memory->rblocks[addr >> 14][addr];
}

uint8 memory_read_byte(memory_o memory, uint addr)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	return _memory->rblocks[addr >> 14][addr];
}

void memory_write_byte(memory_o memory, uint addr, uint8 data)
{
	_memory_o *_memory = (_memory_o *)memory;

	addr &= 0xFFFF;

	_memory->wblocks[addr >> 14][addr] = data;
}

void memory_remap(memory_o memory, uint page, uint addr, uint perm)
{
	_memory_o *_memory = (_memory_o *)memory;
	unsigned char *ptr;

	page &= 0x03;
	addr &= 0x1f;
	if (addr & 0x10)
		addr &= 0x17;

	_memory->pages[page] = addr;

	ptr = _memory->memory + PAGE(addr - page);

	_memory->rblocks[page] = ptr;
	_memory->wblocks[page] = scratch - PAGE(page);

	if (perm & PERM_WRITE)
		_memory->wblocks[page] = ptr;
}

int memory_raw(memory_o memory, void **rd_ptr, void **wr_ptr, void **ex_ptr)
{
	*rd_ptr = NULL;
	*wr_ptr = NULL;
	*ex_ptr = NULL;

	return 0;
}
